import csv
import collections
import pathlib
import argparse
import dataclasses
from typing import List, Dict, Optional

import openpyxl
import openpyxl.styles


DATE = '2022-02-26'
TEST_NAMES = [
    ('test florian', 'test bruhin'),
    ('Test', 'Florian'),
    ('Claudia', 'Furrer'),
    ('Bruhin', 'Florian'),
    ('Test Claudia', 'von Florian'),
]


@dataclasses.dataclass
class Signup:

    lastname: str
    firstname: str
    exams: str
    email: str
    slot: str
    cancelled_reason: Optional[str] = None

    @property
    def is_cancelled(self):
        return self.cancelled_reason is not None

    def _cancelled_str(self):
        assert self.is_cancelled
        if self.cancelled_reason:
            return f'STORNIERT: {self.cancelled_reason}'
        return 'STORNIERT'

    def __str__(self):
        s = f'{self.firstname} {self.lastname} ({self.email}) @ {self.slot}: {self.exams}'
        if self.is_cancelled:
            s += ' - ' + self._cancelled_str()
        return s

    def row_data(self):
        data = [self.lastname, self.firstname, self.exams, self.email]
        if self.is_cancelled:
            data.append(self._cancelled_str())
        return data

    def is_test(self):
        if (self.firstname, self.lastname) in TEST_NAMES:
            assert self.is_cancelled, f"Test booking for {self} is not cancelled"
            return True
        return False


def process(reader: csv.reader) -> Dict[str, List[Signup]]:
    event_data = collections.defaultdict(list)
    all_by_mail = collections.defaultdict(list)
    number_of_custom_questions = sum(
            1 for fieldname in reader.fieldnames if fieldname.startswith('Question ')
    )

    for line in reader:
        date, slot = line['Start Date & Time'].split(' ')
        assert date == DATE, date

        exams = []
        for i in range(1, number_of_custom_questions + 1):
            key = f'Response {i}'
            exams += line[key].split('\n')
        exams = [e for e in exams if e]

        lastname = line['Invitee Last Name']
        firstname = line['Invitee First Name']
        email = line['Invitee Email']

        if not firstname or not lastname:
            print(f"!!! WRONG NAME: {lastname} {firstname} {email} for {slot} {exams}")
        if not exams:
            print(f"!!! NO EXAMS: {lastname} {firstname} {email} for {slot}")
        
        signup = Signup(
            lastname=lastname,
            firstname=firstname,
            exams=' '.join(exams),
            email=email,
            slot=slot,
        )
        if line['Canceled'] == 'true':
            print(f"Cancelled: {lastname} {firstname} for {slot}")
            signup.cancelled_reason = line['Cancellation reason']

        if signup.is_test():
            print(f"DELETING Test: {signup}")
            continue

        all_by_mail[email].append(signup)
        event_data[slot].append(signup)

    print()

    for mail, signups in all_by_mail.items():
        signup_str = '\n' + '\n'.join('    ' + str(s) for s in signups)
        if len(signups) != 1 and len([s for s in signups if not s.is_cancelled]) == 1:
            print(f"multiple (but cancelled) for {mail}: {signup_str}")
            print()
        elif len(signups) != 1:
            print(f"!!! MULTIPLE for {mail}: {signup_str}")
            print()

    return event_data

def _resize_cols(sheet: openpyxl.worksheet.worksheet.Worksheet) -> None:
    dimensions = {
        'A': 20,
        'B': 20,
        'C': 50,
        'D': 30,
        'E': 50,
    }
    for col, dim in dimensions.items():
        sheet.column_dimensions[col].width = dim

def writeout(event_data: Dict[str, List[Signup]], path: pathlib.Path) -> None:
    workbook = openpyxl.Workbook()

    for slot, data in sorted(event_data.items()):
        sheet = workbook.create_sheet(title=slot.replace(':', '.'))
        for signup in data:
            assert signup.slot == slot, signup
            sheet.append(list(signup.row_data()))
            if signup.is_cancelled:
                for tpl in sheet.iter_cols(min_row=sheet.max_row):
                    assert len(tpl) == 1, tpl
                    cell = tpl[0]
                    assert cell.row == sheet.max_row, cell
                    cell.font = openpyxl.styles.Font(
                        color="FF0000",
                        strike=cell.column != sheet.max_column,
                    )

        _resize_cols(sheet)

    del workbook['Sheet']
    workbook.save(path)


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument('path', default='events-export.csv', type=pathlib.Path)
    return parser.parse_args()


def main() -> None:
    args = parse_args()
    with args.path.open() as f:
        reader = csv.DictReader(f)
        event_data = process(reader)
        writeout(event_data, args.path.with_suffix('.xlsx'))


if __name__ == '__main__':
    main()
